# Credits: Built on open-source libraries and papers acknowledged in README.md citations.
#
"""Discriminator architectures for image-to-image translation."""

from src.models.discriminators.patchgan import PatchGANDiscriminator

__all__ = [
    "PatchGANDiscriminator",
]

