# Copyright (c) 2026 EarthBridge Team.
"""DBIM training example for paired image-to-image translation."""

from .config import DBIMConfig
from .train_dbim import DBIMTrainer

__all__ = ["DBIMConfig", "DBIMTrainer"]
