# Copyright (c) 2026 EarthBridge Team.
# Credits: ECSI (https://github.com/szhan311/ECSI), EDM (Karras et al.).
#
"""Re-export EDM from src.models.unet.edm for backward compatibility."""

from src.models.unet.edm import EDM

__all__ = ["EDM"]
