# Copyright (c) 2026 EarthBridge Team.
"""BDBM (Bidirectional Diffusion Bridge Model) training example."""

from .config import BDBMConfig
from .train_bdbm import BDBMTrainer

__all__ = ["BDBMConfig", "BDBMTrainer"]
